private import std.string;
private import std.file;
private import win32.ansi.windows;
private import util;
private import bga_melter;

//----------------------------------------------------------------
// おきまりのDLL初期化ルーチン
//----------------------------------------------------------------

extern(C)
{
	void gc_init();
	void gc_term();
	void _minit();
	void _moduleCtor();
	void _moduleDtor();
	void _moduleUnitTests();
}

extern (Windows)
BOOL DllMain( HINSTANCE inst, ULONG reason, void* reserved )
{
	switch( reason )
	{
	case DLL_PROCESS_ATTACH:
		gc_init();          // GC初期化
		_minit();           // モジュールリスト初期化
		_moduleCtor();      // モジュールコンストラクタ実行
		_moduleUnitTests(); // 単体テスト実行
		if( g_orig_dll is null )
			return false;
		break;

	case DLL_PROCESS_DETACH:
		_moduleDtor();
		gc_term();     // GC終了
		break;

	default:
	case DLL_THREAD_ATTACH:
	case DLL_THREAD_DETACH:
		break;
	}
	return true;
}

//----------------------------------------------------------------
// API転送処理
//----------------------------------------------------------------

HINSTANCE g_orig_dll = null;

static this()
{
	g_orig_dll = LoadLibrary( "_Bga32.DLL" );
}

static ~this()
{
	if( g_orig_dll )
		FreeLibrary( g_orig_dll );
}

template get_api(FnT)
{
	FnT get_api( char[] name )
	{
		return cast(FnT) GetProcAddress( g_orig_dll, name );
	}
}

//----------------------------------------------------------------
// 統合アーカイバAPI：転送
//----------------------------------------------------------------

extern(Windows)
{
	int Bga( HWND a, char* b, char* c, DWORD d )
	{
		int r = Bga_impl( a, toString(b) );
		if( r < 0 ) // このダミーDLLでは処理できないコマンドだった時
			return get_api!(typeof(&Bga))("Bga")(a,b,c,d);
		return r;
	}

	WORD BgaGetVersion()
	{
		return get_api!(typeof(&BgaGetVersion))("BgaGetVersion")();
	}

	BOOL BgaGetRunning()
	{
		return get_api!(typeof(&BgaGetRunning))("BgaGetRunning")();
	}

	BOOL BgaCheckArchive( char* a, int b )
	{
		return get_api!(typeof(&BgaCheckArchive))("BgaCheckArchive")(a,b);
	}

	BOOL BgaConfigDialog( HWND a, char* b, int c )
	{
		return get_api!(typeof(&BgaConfigDialog))("BgaConfigDialog")(a,b,c);
	}

	int BgaGetFileCount( char* a )
	{
		return get_api!(typeof(&BgaGetFileCount))("BgaGetFileCount")(a);
	}

	BOOL BgaQueryFunctionList( int a )
	{
		return get_api!(typeof(&BgaQueryFunctionList))("BgaQueryFunctionList")(a);
	}

	alias void* HARC;
	HARC BgaOpenArchive( HWND a, char* b, DWORD c )
	{
		return get_api!(typeof(&BgaOpenArchive))("BgaOpenArchive")(a,b,c);
	}

	int BgaCloseArchive( HARC a )
	{
		return get_api!(typeof(&BgaCloseArchive))("BgaCloseArchive")(a);
	}

	alias void* LPINDIVIDUALINFO;
	int BgaFindFirst( HARC a, char* b, LPINDIVIDUALINFO c )
	{
		return get_api!(typeof(&BgaFindFirst))("BgaFindFirst")(a,b,c);
	}

	int BgaFindNext( HARC a, LPINDIVIDUALINFO b )
	{
		return get_api!(typeof(&BgaFindNext))("BgaFindNext")(a,b);
	}

	DWORD BgaGetArcOriginalSize( HARC a )
	{
		return get_api!(typeof(&BgaGetArcOriginalSize))("BgaGetArcOriginalSize")(a);
	}

	DWORD BgaGetArcCompressedSize( HARC a )
	{
		return get_api!(typeof(&BgaGetArcCompressedSize))("BgaGetArcCompressedSize")(a);
	}

	WORD BgaGetArcRatio( HARC a )
	{
		return get_api!(typeof(&BgaGetArcRatio))("BgaGetArcRatio")(a);
	}

	BOOL BgaSetOwnerWindow( HWND a )
	{
		return get_api!(typeof(&BgaSetOwnerWindow))("BgaSetOwnerWindow")(a);
	}

	BOOL BgaClearOwnerWindow()
	{
		return get_api!(typeof(&BgaClearOwnerWindow))("BgaClearOwnerWindow")();
	}

	alias void* LPARCHIVERPROC;
	BOOL BgaSetOwnerWindowEx( HWND a, LPARCHIVERPROC b )
	{
		return get_api!(typeof(&BgaSetOwnerWindowEx))("BgaSetOwnerWindowEx")(a,b);
	}

	BOOL BgaKillOwnerWindowEx( HWND a )
	{
		return get_api!(typeof(&BgaKillOwnerWindowEx))("BgaKillOwnerWindowEx")(a);
	}

	alias void* UNLHA_WND_ENUMMEMBPROC;
	BOOL BgaSetEnumMembersProc( UNLHA_WND_ENUMMEMBPROC a )
	{
		return get_api!(typeof(&BgaSetEnumMembersProc))("BgaSetEnumMembersProc")(a);
	}

	BOOL BgaClearEnumMembersProc()
	{
		return get_api!(typeof(&BgaClearEnumMembersProc))("BgaClearEnumMembersProc")();
	}
}

//----------------------------------------------------------------
// 統合アーカイバAPI：実装
//----------------------------------------------------------------

int Bga_impl( HWND wnd, char[] cmd_str )
{
	enum { UNSUPPORTED = -1 }

	//
	// コマンドライン解析
	//
	char[][] cmd = cmd_parse(cmd_str);

	// x以外のコマンドは扱わないで本物DLLに回します。注意点として：
	//  > command はコマンドラインの最初の引数としてください。なお、command を省略
	//  > した場合は 'x' command が指定されたものとみなします。
	if( cmd.length == 0 )
		return UNSUPPORTED;
	if( cmd[0].length == 1 )
	{
		if( 0 <= find("adjlmnstvADJLMNSTV", cmd[0][0]) )
			return UNSUPPORTED;
		if( cmd[0][0]=='x' || cmd[0][0]=='X' )
			cmd = cmd[1 .. length];
	}

		// ※ この時点で、cmdにはcommandを除いた残りの引数が入っているはず

	//
	// スイッチ解析、引数解析
	//
	bool all_attrs       = false; // -a
	bool silent          = false; // -i
	bool ignore_dir      = false; // -j
	bool newfile_only    = false; // -n
	bool force_overwrite = false; // -o
	bool recursive       = false; // -r
	bool sanitize_path   = true;

	char[] arc_name = null;
	char[] base_dir = null;
	char[][] paths;

	foreach( char[] param ; cmd )
		if( param[0] == '-' )
			switch( param[1] )
			{
			case 'a','A': all_attrs       = true; break;
			case 'i','I': silent          = true; break;
			case 'j','J': ignore_dir      = true; break;
			case 'n','N': newfile_only    = true; break;
			case 'o','O': force_overwrite = true; break;
			case 'r','R': recursive       = true; break;
			default:  break;
			}
		else if( arc_name is null )
		{
			arc_name = param;
		}
		else if( base_dir is null )
		{
			if( lastChar(param) == '\\' )
				base_dir = param;
			else {
				base_dir.length = GetCurrentDirectory(0,null)+1;
				GetCurrentDirectory(base_dir.length, base_dir);
				if( lastChar(base_dir) != '\\' )
					base_dir ~= '\\';
			}
		}
		else
			paths ~= param;

	//
	// 展開処理にGo!
	//
	try
	{
		BgaAnswer handler( inout BgaHeader hdr )
		{
			// paths
			if( paths.length > 0 )
			{
				char[] fname = // -r
					(recursive ? hdr.fname[hdr.dir_name_len..length]
					           : hdr.fname);
				foreach( char[] w ; paths )
					if( wild_match( w, fname ) )
						goto ok;
				return BgaAnswer.SkipIt;
			ok:;
			}
			// -a
			if( !all_attrs && (hdr.attrib&6) )
				return BgaAnswer.SkipIt;
			// -j
			if( ignore_dir )
				hdr.fname = hdr.fname[hdr.dir_name_len .. length];
			// sanitize
			if( sanitize_path )
				hdr.fname = check_path(hdr.fname);
			// base_dir
			hdr.fname = base_dir ~ hdr.fname;
			// -o
			if( !force_overwrite )
				try {
				if( std.file.exists(hdr.fname) && std.file.isfile(hdr.fname) )
					// -n
					if( newfile_only )
					{
						if( newer_than(hdr.date,hdr.time,hdr.fname) )
							return BgaAnswer.SkipIt;
					}
					else
					{
						int r = MessageBox( wnd,
							toStringz("Overwrite "~hdr.fname~" ?"),
							"QBga32.dll", MB_YESNOCANCEL );
						if( r == IDNO )     return BgaAnswer.SkipIt;
						if( r == IDCANCEL ) return BgaAnswer.Abort;
					}
				} catch {}
			return BgaAnswer.MeltIt;
		}

		BgaAnswer progress_handler( int cur, int max )
		{
			// 進捗状況表示は今のところなしで…
			return BgaAnswer.MeltIt;
		}

		(new BgaMelter(arc_name)).start(&handler,&progress_handler);
	}
	catch( BgaMelterError e )
	{
		return e.errcode;
	}
	return 0;
}

//----------------------------------------------------------------
// パス検査系
//----------------------------------------------------------------

alias std.c.windows.windows.IsDBCSLeadByte isDL;
char[] replace_yen( char[] s )
{
	char[] ans;
	int j=0;
	for(int i=0; i!=s.length; i=i+(isDL(s[i])?2:1))
		if( s[i] == '\\' )
			ans~=s[j .. i], ans~='/', j=i+1;
	ans ~= s[j .. length];
	return ans;
}

bool wild_match( char[] wild, char[] name )
{
	bool wild_match_nopath( char[] w, char[] s )
	{
		char[] advance( char[] s )
		{
			return s[(IsDBCSLeadByte(s[0])?2:1) .. length];
		}

		while( w.length>0 )
			switch( w[0] )
			{
			case '?':
				if( s.length==0 )
					return false;
				w = advance(w);
				s = advance(s);
				break;
			case '*':
				if( s.length==0 )
					return false;
				w = advance(w);
				if( w.length == 0 )
					return true;
				for( ; s.length!=0; s=advance(s) )
					if( wild_match_nopath(w,s) )
						return true;
				return false;
			default:
				if( s.length==0 )
					return false;
				if( isDL(w[0]) )
				  { if( w[0..2] != s[0..2] ) return false; }
				else
				  { if( w[0]    != s[0]    ) return false; }
				w = advance(w);
				s = advance(s);
				break;
			}
		return s.length==0;
	}

	if( wild=="" || wild=="*.*" || wild=="*" || wild=="**" )
		return true;

	char[][] wilds = split( replace_yen( tolower(wild) ), "/" );
	char[][] names = split( replace_yen( tolower(name) ), "/" );

	if( wilds.length != names.length )
		return false;
	for(int i=0; i!=wilds.length; ++i)
		if( wilds[i]!="*.*" && wilds[i]!="*" && wilds[i]!="**" )
			if( !wild_match_nopath( wilds[i], names[i] ) )
				return false;
	return true;
}

char[] check_path( char[] path )
{
	// C:\ ==> C_\
	if( path.length>=2 && path[1]==':' )
		path = path.dup, path[1] = '_';

	// \\hoge ==> hoge
	// /hoge  ==> hoge
	while( path.length>0 && (path[0]=='\\'||path[0]=='/') )
		path = path[1..length];

	// .. ==> __
	char[][] paths = split( replace_yen(path), "/" );
	L1:
	foreach( inout char[] pc ; paths )
		if( pc.length >= 2 )
		{
			foreach( char c ; pc )
				if( c != '.' )
					continue L1;
			pc = replace( pc, ".", "_" );
		}

	return join( paths, "\\" );
}



//----------------------------------------------------------------
// 簡易テスト＠もっとマジメに書かなきゃ…
//----------------------------------------------------------------

unittest
{
	assert( check_path(`\\\\hoge\fuga`)==`hoge\fuga` );
	assert( check_path(`/usr/local/`)==`usr\local\` );
	assert( check_path(`..\abc def\...\.\g`)==`__\abc def\___\.\g` );
	assert( wild_match(`a/b/c`,`A\b\C`) );
	assert( wild_match(`a/*.*/a?x`,`A\hoge\Afx`) );
	assert( Bga_impl(null,"a hoge") < 0 );
}
